/*
 *    MessageRouter.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.core;

import org.greenstone.gsdl3.util.*;
import org.greenstone.gsdl3.service.*;
import org.greenstone.gsdl3.comms.*;
import org.greenstone.gsdl3.collection.*;

// XML classes
import org.w3c.dom.Node; 
import org.w3c.dom.NodeList; 
import org.w3c.dom.Document; 
import org.w3c.dom.Element; 
import org.xml.sax.InputSource;
import javax.xml.parsers.*;

// other java classes
import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.io.Reader;
import java.io.StringReader;

/**
 * The hub of a Greenstone system. 
 *
 * Accepts  XML requests (via process method of ModuleInterface) and routes them to the appropriate collection or 
 * service or external entity.
 *
 * contains a map of module objects - may be services, collections, comms 
 * objects talking to other MessageRouters etc.
 *
 *
 * @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 * @version $Revision: 3461 $
 * @see ModuleInterface
 * @see Collection
 * @see ServiceModule
 * @see Communicator
 *
 */
public class MessageRouter implements  ModuleInterface {

    // should go in a separate class? do other names too?
    static protected String GSDL_MESSAGE_NAME = "message";

    /** site home */
    protected String site_home_=null;
    /** the global name for this site */
    protected String site_name_=null;

    /** map of name to Module object */
    protected HashMap module_map_=null;

    /** container Document to create XML Nodes */
    protected Document doc_=null; 
    /** the full description of this site */
    protected Element description_ = null;
    /** list of collections that can be reached */
    protected Element collection_list_ = null; 
    /** list of services that can be reached */
    protected Element service_list_ = null; 
    /** list of sites that can be reached */
    protected Element site_list_ = null; 

    /** a converter class to parse XML and create Docs */
    protected XMLConverter converter_=null;

    //***************************************************************
    // public methods
    //***************************************************************

    /** constructor */
    public MessageRouter() {
	module_map_ = new HashMap();
	converter_ = new XMLConverter();
	doc_ = converter_.newDOM();
	
	
    }

    /** site_home_ must be set before configure called */
    public void setSiteHome(String home) {
	site_home_ = home;
    }

    /**
     * configures the system
     *
     * looks in site_home_/collect for collections, reads config file
     * site_home_/sitecfg.xml
     *
     */
    public boolean configure() { 

	if (site_home_==null) {
	    System.err.println("You must set site_home_ before calling configure");
	    return false;
	}

    	// start the description_ XML tree
	description_ = doc_.createElement("response");
	description_.setAttribute("type", "describe");
	collection_list_ = doc_.createElement("collectionList");
	service_list_ = doc_.createElement("serviceList");
	site_list_ = doc_.createElement("siteList");


	// read thru own config file - create services and connect to sites
	File configFile = new File(GSFile.siteConfigFile(site_home_));

	if (configFile.exists() ) {
	    configureSystem(configFile);
	}
	else {
	    System.err.println("MessageRouter: site config file: "+configFile.getPath()+" not found!");
	    return false;
	    
	}

	// read thru the collect directory and create all the collections
	File collectDir = new File(GSFile.collectDir(site_home_));
	if (collectDir.exists()) {
	    System.out.println("Reading thru directory "+collectDir.getPath()+" to find collections.");
	    File[] contents = collectDir.listFiles();
	    for (int i=0; i<contents.length;i++) {
		if(contents[i].isDirectory()) {
		    
		    String colName = contents[i].getName();
		    if (!colName.startsWith("CVS")) {
			// need to ignore CVS
			createCollection(colName);
		    }
		}
	    }
	} // collectDir
	
	description_.appendChild(collection_list_);

	return true;
	
    }

    /**
     * Process an XML request - as a String
     *
     * @param xml_in the request to process
     * @return the response - contains any error messages
     * @see String
     */
    public String process(String xml_in) {
	
	Document doc = converter_.getDOM(xml_in);
	
	Node result = process(doc);
	return converter_.getString(result);
    }
    
    /**
     * Process an XML request - as a DOM model node
     * the Node may be an Element or a Document
     *
     * @param xml_in the message to process - should be <message>
     * @return the response - contains any error messages
     * @see Node
     * @see Document
     * @see Element
     */
    public Node process(Node xml_in) {
	
	// check if Element or Document - we want to deal with Elements
	Element message=null;
	short node_type = xml_in.getNodeType();
	if (node_type == Node.DOCUMENT_NODE) {
	    message = ((Document)xml_in).getDocumentElement();
	} else if (node_type == Node.ELEMENT_NODE) {
	    message = (Element)xml_in;
	} else {
	    System.err.println("MessageRouter:process: invalid Node type");
	}

	// check that its a correct message tag
	if (!message.getTagName().equals(GSDL_MESSAGE_NAME)) {
	    System.out.println("Invalid message. GSDL message should start with <"+GSDL_MESSAGE_NAME+">, instead it starts with:"+message.getTagName()+".");
	    return null;
	}

	NodeList requests = message.getElementsByTagName("request");

	Element mainResult = doc_.createElement("message");
	
	// empty request
	if (requests.getLength()==0) {
	    return mainResult;
	}

	// for now, just process each request one by one
	for (int i=0; i<requests.getLength(); i++) {
	    Node result=null;
	    Element req = (Element)requests.item(i);
	    String path = req.getAttribute("to"); // returns "" if no att of this name
	    if (path.equals("")) {
		// its a message for the message router
		result = processMessage(req);
	    } else {
		// find the module to pass it on to
		String obj = GSPath.getFirstLink(path);
		if (module_map_.containsKey(obj)) {
		    result = ((ModuleInterface)module_map_.get(obj)).process(req);
		} else {    
		    System.err.println("MessageRouter: request has illegal module name in:\n"+converter_.getString(req));
		}
	    }
	    
	    if (result !=null ) {
		// hack for now for external comms stuff
		if (result.getNodeName().equals("message")) {
		    // remove the message bit
		    Node child = result.getFirstChild();
		    while (child !=null) {
			mainResult.appendChild(doc_.importNode(child, true));
			child = child.getNextSibling();
		    }
		} else {
		    // just append the whole node
		    mainResult.appendChild(doc_.importNode(result, true));
		}
	    }
	} // for each request
	return mainResult;
	
    }
	
    //********************************************************************
    // auxiliary configure methods
    // *******************************************************************

    /**
     * configures the message router based on the site config file 
     * - sitecfg.xml
     *
     */
    protected boolean configureSystem(File file) {

	System.out.println("MessageRouter:configuring site");
	Document config_doc = converter_.getDOM(file);

	NodeList local_site_name = config_doc.getElementsByTagName("localSiteName");
	if (local_site_name.getLength()==0) {
	    System.err.println("MessageRouter configure error:no site name in config file");
	    return false;
	} else {
	    Element elem = (Element)local_site_name.item(0);
	    site_name_ = elem.getAttribute("value");
	}

	// load up the services
	System.out.println("loading service modules...");

	NodeList service_modules = config_doc.getElementsByTagName("serviceModule");
	if (service_modules.getLength()==0) {
	    System.out.println("... none to be loaded");
	} else {
	
	    Element service_request = doc_.createElement("request");
	    service_request.setAttribute("type", "describe");
	    service_request.setAttribute("info", "serviceList");
	    
	    for(int i=0; i<service_modules.getLength(); i++) {
		Element n = (Element)service_modules.item(i);
		String service_name = n.getAttribute("name");
		System.out.println("..."+service_name);
		try {
		    ServiceModule s = (ServiceModule)Class.forName("org.greenstone.gsdl3.service."+service_name).newInstance();
		    s.setSiteHome(site_home_);
		    s.setMessageRouter(this);
		    
		    // pass the XML node to the service for service configuration
		    s.configure(n);
		    
		    // find out the supported services for this service module
		    Element service_response = (Element) s.process(service_request);
		    NodeList services = service_response.getElementsByTagName("service");
		    if (services.getLength()==0) {
			System.err.println("MessageRouter configure error: serviceModule "+service_name+" has no services!");
		    } else {
			for (int j=0; j<services.getLength();j++) {
			    String service = ((Element)services.item(j)).getAttribute("name");
			module_map_.put(service, s);
			
			// add short info to service_list_ XML
			service_list_.appendChild(doc_.importNode(services.item(j), true));
			}
		    }
		} catch (Exception e ) {
		    System.err.println("MessageRouter configure exception:  in ServiceModule class specification:  "+ e.getMessage());
		    
		} 
		
	    } // for each service module
	}
	
	// load up the sites
	System.out.println("loading external sites...");
	NodeList sites = config_doc.getElementsByTagName("site");
	if (sites.getLength()==0) {
	    System.out.println("...none found");
	} else {

	    for (int i=0; i<sites.getLength(); i++) {
		Element s = (Element)sites.item(i);
		Communicator comm=null;
		String type = s.getAttribute("type");
		String name = s.getAttribute("name");
		System.out.println("..."+name+", type="+type);
		 if (type.equals("soap")) {
		    comm = new SOAPCommunicator();
		    comm.setSiteHome(site_home_);
		    comm.setConnectionName(name);
		    comm.setConnectionAddress(s.getAttribute("address"));
		    comm.setSiteName(site_name_);
		    
		    // add to map of modules
		    module_map_.put(name, comm);
		    site_list_.appendChild(doc_.importNode(s, true));
		    // need to get collection list and service list from here
		    // this may introduce a configure problem - leave as 
		    // is for now, but fix up later - kjdon
		    getRemoteCollectionInfo(comm, name);
		    getRemoteServiceInfo(comm, name);

		} else {
		    System.err.print("cant talk to server of type:"+type);
		    System.err.println(", so not making a connection to "+name);
		}
		
		    
	    }
	}
	// add children to description_
	description_.appendChild(service_list_);
	description_.appendChild(site_list_);
		
	return true;

    }

    /** get collection list from external site
     *
     * @param comm - the communicator object for the external site
     * @param site_name - the name of the external site
     * @return true if successful
     */
    protected boolean getRemoteCollectionInfo(Communicator comm, String site_name) {
	System.out.println("MessageRouter: getting coll info from site:"+site_name);
	String coll_request = "<request type='describe' info='collectionList'/>";

	String coll_response = comm.process(coll_request);
	
	Document coll_doc = converter_.getDOM(coll_response);
	NodeList colls = coll_doc.getElementsByTagName("collection");
	if (colls.getLength()>0) {
	    for (int i=0; i<colls.getLength(); i++) {
		Element e = (Element)colls.item(i);
		String col_name = e.getAttribute("name");
		// add the info to own coll list - may want to keep 
		// this separate in future - so can distinguish own and 
		// other collections
		e.setAttribute("name", GSPath.prependLink(col_name, site_name));
		collection_list_.appendChild(doc_.importNode(e, true));
	    }
	}
	return true;
		    
    }
    /** get service list from external site
     *
     * @param comm - the communicator object for the external site
     * @param site_name - the name of the external site
     * @return true if successful
     */
    protected boolean getRemoteServiceInfo(Communicator comm, String site_name) {
	System.out.println("MessageRouter: getting service info from site:"+site_name);
	String serv_request = "<request type='describe' info='serviceList'/>";
	String serv_response = comm.process(serv_request);
	Document serv_doc = converter_.getDOM(serv_response);
	NodeList services = serv_doc.getElementsByTagName("service");
	if (services.getLength()>0) {
	    for (int i=0; i<services.getLength(); i++) {
		Element e = (Element)services.item(i);
		String serv_name = e.getAttribute("name");
		System.out.println("found service name="+serv_name);
		e.setAttribute("name", GSPath.prependLink(serv_name, site_name));
		service_list_.appendChild(doc_.importNode(e, true));
	    }
	}
	return true;

    }

    /** creates and configures a new collection
     *
     *@param col_name the name of the collection
     *@return true if collection created ok
     */
    protected boolean createCollection(String col_name) {

	System.out.println("MessageRouter:Creating and configuring collection: "+col_name+".");
			  
	Collection c = new Collection();
	c.setCollectionName(col_name);
	c.setSiteHome(site_home_);
	if (c.configure()) {
	    // add to list of collections
	    module_map_.put(col_name, c);
	    //add short description_ to collection_list_
	    Element e = doc_.createElement("collection");
	    e.setAttribute("name", col_name);
	    collection_list_.appendChild(e);
	    return true;
	} else {
	    System.out.println("MessageRouter:Couldn't configure collection: "+
			       col_name+".");
	    return false;
	}
	
    }

    //*****************************************************************
    // auxiliary process methods
    //*****************************************************************

    /** handles requests made to the MessageRouter itself 
     *
     * @param req - the request Element- <request>
     * @return the result Element - should be <response>
     */
    protected Element processMessage(Element req) {

	// message for self, should be type=describe at this stage
	String type = req.getAttribute("type");
	
	if (type.equals("describe")) {
	    String info = req.getAttribute("info");
	    if (info.equals("")) {
		// return full description
		return description_;
	    }
	    // else create a new response with the appropriate info 
	    Element response = doc_.createElement("response");
	    response.setAttribute("type", "describe");
	    
	    if (info.equals("collectionList")) {
		response.appendChild(collection_list_);
		return response;
	    } else if (info.equals("serviceList")) {
		response.appendChild(service_list_);
		return response;
	    } else if (info.equals("sitelist")) {
		response.appendChild(site_list_);
		return response;
	    } 
	    
	}
	// if get here something has gone wrong
	System.err.println("MessageRouter: cant process request:");
	System.err.println(converter_.getString(req));
	return null;
       
    }

}
